/*
 *  NI-Serial PXI(e) Driver
 *
 *  This module driver provides functionality support for
 *  National Instruments (NI) NI-Serial hardware
 *  and uses the 8250 driver for the UART functionality.
 *
 *  Copyright 2018 National Instruments Corporation
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 *  more details.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/8250_pci.h>
#include <linux/serial_8250.h>
#include <linux/serial_core.h>
#include <linux/serial_reg.h>
#include "niserial_pci_ids.h"

#define NISERIAL_MAX_NUM_PORT  16
struct niserial_private {
	void __iomem *p;
	struct pci_dev *dev;
	unsigned int nr;
	int line[NISERIAL_MAX_NUM_PORT];
};

struct niserial_board {
	unsigned int num_ports;
	unsigned int uartclk;
	void   (*setup)(void __iomem *,
			struct uart_8250_port *);
};

#define MITE_IOWBSR1_WSIZE	0xa
#define MITE_IOWBSR1_WIN_OFFSET	0x800
#define MITE_IOWBSR1_WENAB	(1 << 7)
#define MITE_LCIMR1_IO_IE_0	(1 << 24)
#define MITE_LCIMR2_CLR_CPU_IE (1 << 30)
#define MITE_LCIMR2_SET_CPU_IE	(1 << 31)
#define MITE_IOWCR1_RAMSEL_MASK	0xfffffffe
#define MITE_IOWBSR1 0xc4
#define MITE_IOWCR1 0xf4
#define MITE_LCIMR1 0x08
#define MITE_LCIMR2 0x10

static void pci_ni8430_init(struct pci_dev *dev, struct niserial_private *priv)
{
	struct pci_bus_region region;
	u32 device_window;

	/*
	* Set device window address and size in BAR0, while acknowledging that
	* the resource structure may contain a translated address that differs
	* from the address the device responds to.
	*/
	pcibios_resource_to_bus(dev->bus, &region, &dev->resource[0]);
	device_window = ((region.start + MITE_IOWBSR1_WIN_OFFSET) & 0xffffff00)
			| MITE_IOWBSR1_WENAB | MITE_IOWBSR1_WSIZE;
	writel(device_window, priv->p + MITE_IOWBSR1);

	/* Set window access to go to RAMSEL IO address space */
	writel((readl(priv->p + MITE_IOWCR1) & MITE_IOWCR1_RAMSEL_MASK),
		priv->p + MITE_IOWCR1);

	/* Enable IO Bus Interrupt 0 */
	writel(MITE_LCIMR1_IO_IE_0, priv->p + MITE_LCIMR1);

	/* Enable CPU Interrupt */
	writel(MITE_LCIMR2_SET_CPU_IE, priv->p + MITE_LCIMR2);
}

#define NI16550_UART_OFFSET 0x10
#define NI16550_UART_FIRST_OFFSET 0x800

/*
 * Probe the serial board.
 */
static int niserial_pci_probe(struct pci_dev *dev,
	const struct pci_device_id *ent)
{
	const struct niserial_board *board;
	void __iomem *p;
	struct niserial_private *priv;
	struct uart_8250_port uart;
	struct resource *res = &dev->resource[0];
	int  i, rc, nr_ports;
	unsigned int offset = NI16550_UART_FIRST_OFFSET;

	board = (const struct niserial_board *)ent->driver_data;

	rc = pcim_enable_device(dev);
	if (rc)
		return rc;

	nr_ports = board->num_ports;

	p = devm_ioremap(&dev->dev, res->start, resource_size(res));
	if (!p)
		return -ENOMEM;

	priv = devm_kzalloc(&dev->dev, sizeof(struct niserial_private),
		GFP_KERNEL);

	if (!priv)
		return -ENOMEM;

	priv->dev = dev;
	priv->p = p;

	pci_ni8430_init(dev, priv);

	memset(&uart, 0, sizeof(uart));
	uart.port.flags = UPF_SKIP_TEST | UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ;
	uart.port.uartclk = board->uartclk;
	uart.port.irq = dev->irq;
	uart.port.dev = &dev->dev;
	uart.port.iotype = UPIO_MEM;
	uart.port.iobase = 0;
	uart.port.regshift = 0;

	for (i = 0; i < nr_ports; i++) {

		board->setup(priv->p + offset, &uart);

		uart.port.mapbase = res->start + offset;
		uart.port.membase = priv->p + offset;

		dev_dbg(&dev->dev, "Setup PCI port: port %lx, irq %d, type %d\n",
			uart.port.iobase, uart.port.irq, uart.port.iotype);

		priv->line[i] = serial8250_register_8250_port(&uart);
		if (priv->line[i] < 0) {
			dev_err(&dev->dev,
				"Couldn't register serial port %lx, irq %d, type %d, error %d\n",
				uart.port.iobase, uart.port.irq,
				uart.port.iotype, priv->line[i]);
			break;
		}
		offset += NI16550_UART_OFFSET;
	}
	priv->nr = i;

	pci_set_drvdata(dev, priv);
	return 0;
}

/* UART Port Control Register */
#define NI16550_PCR_OFFSET 0x0f
#define NI16550_PCR_RS422 0x00
#define NI16550_PCR_ECHO_RS485 0x01
#define NI16550_PCR_DTR_RS485 0x02
#define NI16550_PCR_AUTO_RS485 0x03
#define NI16550_PCR_WIRE_MODE_MASK 0x03
#define NI16550_PCR_TXVR_ENABLE_BIT (1 << 3)
#define NI16550_PCR_RS485_TERMINATION_BIT (1 << 6)
#define NI16550_ACR_DTR_AUTO_DTR (0x2 << 3)
#define NI16550_ACR_DTR_MANUAL_DTR (0x0 << 3)

static void niserial_pci_remove(struct pci_dev *dev)
{
	struct niserial_private *priv = pci_get_drvdata(dev);
	int i;
	unsigned int offset = NI16550_UART_FIRST_OFFSET;

	for (i = 0; i < priv->nr; i++) {
		serial8250_unregister_port(priv->line[i]);

		/* Disable the transceiver */
		writeb(readb(priv->p + offset + NI16550_PCR_OFFSET) &
			~NI16550_PCR_TXVR_ENABLE_BIT,
			priv->p + offset + NI16550_PCR_OFFSET);

		offset += NI16550_UART_OFFSET;
	}

	/* Disable the CPU Interrupt */
	writel(MITE_LCIMR2_CLR_CPU_IE, priv->p + MITE_LCIMR2);
}

static int pci_ni8431_config_rs485(struct uart_port *port,
	struct serial_rs485 *rs485)
{
	u8 pcr, acr;

	struct uart_8250_port *up =
		container_of(port, struct uart_8250_port, port);
	acr = up->acr;

	dev_dbg(port->dev, "ni16550_config_rs485\n");

	/* "rs485" should be given to us non-NULL. */
	BUG_ON(rs485 == NULL);

	pcr = port->serial_in(port, NI16550_PCR_OFFSET);
	pcr &= ~NI16550_PCR_WIRE_MODE_MASK;

	if (rs485->flags & SER_RS485_ENABLED) {
		/* RS-485 */
		if ((rs485->flags & SER_RS485_RX_DURING_TX) &&
			(rs485->flags & SER_RS485_RTS_ON_SEND)) {
			dev_dbg(port->dev, "Invalid 2-wire mode\n");
			return -EINVAL;
		}

		if (rs485->flags & SER_RS485_RX_DURING_TX) {
			/* Echo */
			dev_vdbg(port->dev, "2-wire DTR with echo\n");
			pcr |= NI16550_PCR_ECHO_RS485;
			acr |= NI16550_ACR_DTR_MANUAL_DTR;
		} else {
			/* Auto or DTR */
			if (rs485->flags & SER_RS485_RTS_ON_SEND) {
				/* Auto */
				dev_vdbg(port->dev, "2-wire Auto\n");
				pcr |= NI16550_PCR_AUTO_RS485;
				acr |= NI16550_ACR_DTR_AUTO_DTR;
			} else {
				/* DTR-controlled */
				/* No Echo */
				dev_vdbg(port->dev, "2-wire DTR no echo\n");
				pcr |= NI16550_PCR_DTR_RS485;
				acr |= NI16550_ACR_DTR_MANUAL_DTR;
			}
		}
	} else {
		/* RS-422 */
		dev_vdbg(port->dev, "4-wire\n");
		pcr |= NI16550_PCR_RS422;
		acr |= NI16550_ACR_DTR_MANUAL_DTR;
	}

	dev_dbg(port->dev, "write pcr: 0x%08x\n", pcr);
	port->serial_out(port, NI16550_PCR_OFFSET, pcr);

	up->acr = acr;
	port->serial_out(port, UART_SCR, UART_ACR);
	port->serial_out(port, UART_ICR, up->acr);

	/* Update the cache. */
	port->rs485 = *rs485;

	dev_dbg(port->dev, "ni16550_config_rs485\n");
	return 0;
}

static void pci_ni8430_setup(void __iomem *addr, struct uart_8250_port *port)
{
	/* enable the transceiver */
	writeb(readb(addr + NI16550_PCR_OFFSET) | NI16550_PCR_TXVR_ENABLE_BIT,
		addr + NI16550_PCR_OFFSET);
}

static void pci_ni8431_setup(void __iomem *addr, struct uart_8250_port *uart)
{
	u8 pcr, acr;

	/* enable the transceiver */
	writeb(readb(addr + NI16550_PCR_OFFSET) | NI16550_PCR_TXVR_ENABLE_BIT,
		addr + NI16550_PCR_OFFSET);

	pcr = readb(addr + NI16550_PCR_OFFSET);
	pcr &= ~NI16550_PCR_WIRE_MODE_MASK;

	/* set wire mode to default RS-422 */
	pcr |= NI16550_PCR_RS422;
	acr = NI16550_ACR_DTR_MANUAL_DTR;

	/* write port configuration to register */
	writeb(pcr, addr + NI16550_PCR_OFFSET);

	/* access and write to UART acr register */
	writeb(UART_ACR, addr + UART_SCR);
	writeb(acr, addr + UART_ICR);

	uart->port.rs485_config = &pci_ni8431_config_rs485;
}

#define NISERIAL_DEVICE(devid, bd) {               \
	PCI_DEVICE_SUB(                                \
		PCI_VENDOR_ID_NI,                          \
		devid,                                     \
		PCI_ANY_ID,                                \
		PCI_ANY_ID), 0, 0,	                       \
		(kernel_ulong_t)&bd                        \
}

static const struct niserial_board pbn_ni_pcie8430_16 = {
	.num_ports	= 16,
	.uartclk	= 3125000 * 16,
	.setup		= pci_ni8430_setup,
};

static const struct niserial_board pbn_ni_pcie8430_8 = {
	.num_ports	= 8,
	.uartclk	= 3125000 * 16,
	.setup		= pci_ni8430_setup,
};

static const struct niserial_board pbn_ni_pci8431_2 = {
	.num_ports	= 2,
	.uartclk	= 3686400 * 16,
	.setup		= pci_ni8431_setup,
};

static const struct niserial_board pbn_ni_pci8431_4 = {
	.num_ports	= 4,
	.uartclk	= 3686400 * 16,
	.setup		= pci_ni8431_setup,
};


static const struct niserial_board pbn_ni_pci8431_8 = {
	.num_ports	= 8,
	.uartclk	= 3686400 * 16,
	.setup		= pci_ni8431_setup,
};

static const struct niserial_board pbn_ni_pcie8431_8 = {
	.num_ports	= 8,
	.uartclk	= 3125000 * 16,
	.setup		= pci_ni8431_setup,
};

static const struct niserial_board pbn_ni_pcie8431_16 = {
	.num_ports	= 16,
	.uartclk	= 3125000 * 16,
	.setup		= pci_ni8431_setup,
};

static struct pci_device_id niserial_pci_tbl[] = {
	/* NI PXIe RS-232 Devices */
	NISERIAL_DEVICE(PCIE_DEVICE_ID_NI_PXIE8430_2328, pbn_ni_pcie8430_8),
	NISERIAL_DEVICE(PCIE_DEVICE_ID_NI_PXIE8430_23216, pbn_ni_pcie8430_16),
	/* NI PXI(e) RS-485 Devices */
	NISERIAL_DEVICE(PCI_DEVICE_ID_NI_PXI8431_4852, pbn_ni_pci8431_2),
	NISERIAL_DEVICE(PCI_DEVICE_ID_NI_PXI8431_4854, pbn_ni_pci8431_4),
	NISERIAL_DEVICE(PCI_DEVICE_ID_NI_PXI8431_4858, pbn_ni_pci8431_8),
	NISERIAL_DEVICE(PCI_DEVICE_ID_NI_PXI8433_4852, pbn_ni_pci8431_2),
	NISERIAL_DEVICE(PCI_DEVICE_ID_NI_PXI8433_4854, pbn_ni_pci8431_4),
	NISERIAL_DEVICE(PCIE_DEVICE_ID_NI_PXIE8431_4858, pbn_ni_pcie8431_8),
	NISERIAL_DEVICE(PCIE_DEVICE_ID_NI_PXIE8431_48516, pbn_ni_pcie8431_16),
	{0,}
};

static struct pci_driver niserial_pci_driver = {
	.name		= "niserial",
	.id_table = niserial_pci_tbl,
	.probe = niserial_pci_probe,
	.remove = niserial_pci_remove,
};

module_pci_driver(niserial_pci_driver);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("NI-Serial PXI(e) Driver");
MODULE_DEVICE_TABLE(pci, niserial_pci_tbl);
MODULE_AUTHOR("Stephen Lee <stephen.lee@ni.com>");
MODULE_VERSION("18.1");
